"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const fs_1 = tslib_1.__importDefault(require("fs"));
const path_1 = tslib_1.__importDefault(require("path"));
const ApiEmitter_1 = tslib_1.__importDefault(require("./ApiEmitter"));
const FacebookDeviceId_1 = tslib_1.__importDefault(require("./FacebookDeviceId"));
const FacebookHttpApi_1 = tslib_1.__importDefault(require("./FacebookHttpApi"));
const MqttApi_1 = tslib_1.__importDefault(require("./mqtt/MqttApi"));
const PlainFileTokenStorage_1 = tslib_1.__importDefault(require("./PlainFileTokenStorage"));
const debug_1 = tslib_1.__importDefault(require("debug"));
const debugLog = debug_1.default("fblib");
// 🥖
class FacebookApi {
    constructor(options = { selfListen: false }) {
        this.seqId = "";
        this.mqttApi = new MqttApi_1.default();
        this.httpApi = new FacebookHttpApi_1.default();
        const storage = new PlainFileTokenStorage_1.default();
        let session = storage.readSession();
        if (!session) {
            session = { tokens: null, deviceId: null };
        }
        if (!session.deviceId) {
            const deviceId = FacebookDeviceId_1.default();
            session.deviceId = deviceId;
            storage.writeSession({ deviceId, tokens: null });
        }
        if (session.tokens) {
            this.httpApi.token = session.tokens.access_token;
        }
        this.session = session;
        this.emitter = new ApiEmitter_1.default(options, session);
    }
    on(event, callback) {
        this.emitter.on(event, callback);
    }
    once(event, callback) {
        this.emitter.once(event, callback);
    }
    doLogin(login, password) {
        return new Promise((resolve, reject) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (!this.session.tokens) {
                const tokens = yield this.httpApi.auth(login, password);
                this.httpApi.token = tokens.access_token;
                this.session.tokens = tokens;
                const storage = new PlainFileTokenStorage_1.default();
                storage.writeSession(this.session);
            }
            this.mqttApi.on("publish", (publish) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                debugLog(publish.topic);
                if (publish.topic === "/t_ms")
                    this.handleMS(publish.content.toString("utf8"));
            }));
            this.mqttApi.on("connected", () => tslib_1.__awaiter(this, void 0, void 0, function* () {
                const { viewer } = yield this.httpApi.querySeqId();
                const seqId = viewer.message_threads.sync_sequence_id;
                this.seqId = seqId;
                resolve();
                if (!this.session.tokens.syncToken) {
                    yield this.createQueue(seqId);
                    return;
                }
                yield this.createQueue(seqId);
            }));
            yield this.mqttApi.connect();
            yield this.mqttApi.sendConnectMessage(this.session.tokens, this.session.deviceId);
        }));
    }
    sendMessage(threadId, message) {
        return this.mqttApi.sendMessage(threadId, message);
    }
    sendAttachmentFile(to, attachmentPath) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (!fs_1.default.existsSync(attachmentPath))
                throw new Error('Attachment missing!');
            const stream = fs_1.default.createReadStream(attachmentPath);
            const extension = path_1.default.parse(attachmentPath).ext;
            this.httpApi.sendImage(stream, extension, this.session.tokens.uid, to);
        });
    }
    sendAttachmentStream(to, extension, attachment) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            this.httpApi.sendImage(attachment, extension, this.session.tokens.uid, to);
        });
    }
    getThreadInfo(threadId) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const res = yield this.httpApi.threadQuery(threadId);
            const thread = res[threadId];
            const customizations = thread.customization_info;
            return {
                id: thread.thread_key.thread_fbid || thread.thread_key.other_user_id,
                name: thread.name,
                isGroup: thread.is_group_thread,
                participants: thread.all_participants.nodes
                    .map(user => user.messaging_actor)
                    .map(this.parseUser),
                image: thread.image,
                unreadCount: thread.unread_count,
                canReply: thread.can_viewer_reply,
                cannotReplyReason: thread.cannot_reply_reason,
                isArchived: thread.has_viewer_archived,
                color: customizations && customizations.outgoing_bubble_color ?
                    '#' + customizations.outgoing_bubble_color.substr(2).toLowerCase() : null,
                emoji: customizations ? customizations.custom_like_emoji : null,
            };
        });
    }
    getUserInfo(userId) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return this.parseUser((yield this.httpApi.userQuery(userId))[userId]);
        });
    }
    parseUser(user) {
        return {
            id: user.id,
            name: user.name,
            type: user.__type__.name,
            canMessage: user.can_viewer_message,
            emailAddresses: user.email_addresses,
            isBlocked: user.is_blocked_by_viewer,
            isMessengerUser: user.is_messenger_user,
            isPage: user.is_commerce,
            profilePicLarge: user.profile_pic_large ? user.profile_pic_large.uri : null,
            profilePicMedium: user.profile_pic_medium ? user.profile_pic_medium.uri : null,
            profilePicSmall: user.profile_pic_small ? user.profile_pic_small.uri : null
        };
    }
    createQueue(seqId) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const obj = {
                delta_batch_size: 125,
                max_deltas_able_to_process: 1250,
                sync_api_version: 3,
                encoding: "JSON",
                initial_titan_sequence_id: seqId,
                device_id: this.session.deviceId.deviceId,
                entity_fbid: this.session.tokens.uid,
                queue_params: {
                    buzz_on_deltas_enabled: "false",
                    graphql_query_hashes: {
                        xma_query_id: "10153919431161729"
                    },
                    graphql_query_params: {
                        "10153919431161729": {
                            xma_id: "<ID>"
                        }
                    }
                }
            };
            yield this.mqttApi.sendPublish("/messenger_sync_create_queue", JSON.stringify(obj));
        });
    }
    connectQueue(seqId) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const obj = {
                delta_batch_size: 125,
                max_deltas_able_to_process: 1250,
                sync_api_version: 3,
                encoding: "JSON",
                last_seq_id: seqId,
                sync_token: this.session.tokens.syncToken
            };
            yield this.mqttApi.sendPublish("/messenger_sync_get_diffs", JSON.stringify(obj));
        });
    }
    handleMS(ms) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const data = JSON.parse(ms.replace("\u0000", ""));
            // Handled on queue creation
            if (data.syncToken) {
                this.session.tokens.syncToken = data.syncToken;
                const storage = new PlainFileTokenStorage_1.default();
                yield storage.writeSession(this.session);
                yield this.connectQueue(this.seqId);
                return;
            }
            if (!data.deltas || !data.deltas.length)
                return;
            const event = data.deltas[0];
            debugLog(event);
            this.emitter.handleMessage(event);
        });
    }
}
exports.default = FacebookApi;
//# sourceMappingURL=FacebookApi.js.map